# Tests construct several MockWidget classes with different signals
# pylint: disable=function-redefined
# pylint: disable=unused-variable

# It is our constitutional right to use foo and bar as fake handler names
# pylint: disable=blacklisted-name

import unittest
from unittest.mock import patch, MagicMock

from orangewidget.widget import \
    Single, Multiple, Default, NonDefault, Explicit, Dynamic
from orangewidget.tests.base import GuiTest
from orangewidget.utils.signals import _Signal, Input, Output, \
    WidgetSignalsMixin, InputSignal, OutputSignal, MultiInput, summarize, \
    PartialSummary
from orangewidget.widget import OWBaseWidget


class SignalTest(unittest.TestCase):
    def test_get_flags(self):
        self.assertEqual(_Signal.get_flags(False, False, False, False),
                         Single | NonDefault)
        self.assertEqual(_Signal.get_flags(True, False, False, False),
                         Multiple | NonDefault)
        self.assertEqual(_Signal.get_flags(False, True, False, False),
                         Single | Default)
        self.assertEqual(_Signal.get_flags(False, False, True, False),
                         Single | NonDefault | Explicit)
        self.assertEqual(_Signal.get_flags(False, False, False, True),
                         Single | NonDefault | Dynamic)


class InputTest(unittest.TestCase):
    def test_init(self):
        with patch("orangewidget.utils.signals._Signal.get_flags",
                   return_value=42) as getflags:
            signal = Input("a name", int, "an id", "a doc", ["x"])
            self.assertEqual(signal.name, "a name")
            self.assertEqual(signal.type, int)
            self.assertEqual(signal.id, "an id")
            self.assertEqual(signal.doc, "a doc")
            self.assertEqual(signal.replaces, ["x"])
            self.assertEqual(signal.flags, 42)
            getflags.assert_called_with(False, False, False, False)

            Input("a name", int, "an id", "a doc", ["x"], multiple=True)
            getflags.assert_called_with(True, False, False, False)

            Input("a name", int, "an id", "a doc", ["x"], default=True)
            getflags.assert_called_with(False, True, False, False)

            Input("a name", int, "an id", "a doc", ["x"], explicit=True)
            getflags.assert_called_with(False, False, True, False)

    def test_decorate(self):
        input = Input("a name", int)
        self.assertEqual(input.handler, "")

        @input
        def foo():
            pass
        self.assertEqual(input.handler, "foo")

        with self.assertRaises(ValueError):
            @input
            def bar():
                pass


class OutputTest(unittest.TestCase):
    def test_init(self):
        with patch("orangewidget.utils.signals._Signal.get_flags",
                   return_value=42) as getflags:
            signal = Output("a name", int, "an id", "a doc", ["x"])
            self.assertEqual(signal.name, "a name")
            self.assertEqual(signal.type, int)
            self.assertEqual(signal.id, "an id")
            self.assertEqual(signal.doc, "a doc")
            self.assertEqual(signal.replaces, ["x"])
            self.assertEqual(signal.flags, 42)
            getflags.assert_called_with(False, False, False, True)

            Output("a name", int, "an id", "a doc", ["x"], default=True)
            getflags.assert_called_with(False, True, False, True)

            Output("a name", int, "an id", "a doc", ["x"], explicit=True)
            getflags.assert_called_with(False, False, True, True)

            Output("a name", int, "an id", "a doc", ["x"], dynamic=False)
            getflags.assert_called_with(False, False, False, False)

    def test_bind_and_send(self):
        widget = MagicMock()
        output = Output("a name", int, "an id", "a doc", ["x"])
        bound = output.bound_signal(widget)
        value = object()
        id = 42
        bound.send(value, id)
        widget.signalManager.send.assert_called_with(
            widget, "a name", value, id)


class WidgetSignalsMixinTest(GuiTest):
    def test_init_binds_outputs(self):
        class MockWidget(OWBaseWidget):
            name = "foo"
            class Outputs:
                an_output = Output("a name", int)

        widget = MockWidget()
        self.assertEqual(widget.Outputs.an_output.widget, widget)
        self.assertIsNone(MockWidget.Outputs.an_output.widget)

    def test_checking_invalid_inputs(self):
        with self.assertRaises(ValueError):
            class MockWidget(OWBaseWidget):
                name = "foo"

                class Inputs:
                    an_input = Input("a name", int)

        with self.assertRaises(ValueError):
            class MockWidget(OWBaseWidget):
                name = "foo"
                inputs = [("a name", int, "no_such_handler")]

        # Now, don't crash
        class MockWidget(OWBaseWidget):
            name = "foo"
            inputs = [("a name", int, "handler")]

            def handler(self):
                pass

    def test_signal_conversion(self):
        class MockWidget(OWBaseWidget):
            name = "foo"
            inputs = [("name 1", int, "foo"), InputSignal("name 2", int, "foo")]
            outputs = [("name 3", int), OutputSignal("name 4", int)]

            def foo(self):
                pass

        input1, input2 = MockWidget.inputs
        self.assertIsInstance(input1, InputSignal)
        self.assertEqual(input1.name, "name 1")
        self.assertIsInstance(input2, InputSignal)
        self.assertEqual(input2.name, "name 2")

        output1, output2 = MockWidget.outputs
        self.assertIsInstance(output1, OutputSignal)
        self.assertEqual(output1.name, "name 3")
        self.assertIsInstance(output2, OutputSignal)
        self.assertEqual(output2.name, "name 4")

    def test_get_signals(self):
        class MockWidget(OWBaseWidget):
            name = "foo"
            inputs = [("a name", int, "foo")]
            outputs = [("another name", float)]

            def foo(self):
                pass

        self.assertIs(MockWidget.get_signals("inputs"), MockWidget.inputs)
        self.assertIs(MockWidget.get_signals("outputs"), MockWidget.outputs)

        class MockWidget(OWBaseWidget):
            name = "foo"

            class Inputs:
                an_input = Input("a name", int)

            class Outputs:
                an_output = Output("another name", int)

            @Inputs.an_input
            def foo(self):
                pass

        input, = MockWidget.get_signals("inputs")
        self.assertIsInstance(input, InputSignal)
        self.assertEqual(input.name, "a name")

        output, = MockWidget.get_signals("outputs")
        self.assertIsInstance(output, OutputSignal)
        self.assertEqual(output.name, "another name")

    def test_get_signals_order(self):
        class TestWidget(WidgetSignalsMixin):
            class Inputs:
                input_1 = Input("1", int)
                input_2 = Input("2", int)
                input_3 = Input("3", int)
                input_a = Input("a", object)

            class Outputs:
                output_1 = Output("1", int)
                output_2 = Output("2", int)
                output_3 = Output("3", int)
                output_a = Output("a", object)

        inputs = TestWidget.get_signals("inputs")
        self.assertTrue(all(isinstance(s, Input) for s in inputs))
        self.assertSequenceEqual([s.name for s in inputs], list("123a"))
        outputs = TestWidget.get_signals("outputs")
        self.assertTrue(all(isinstance(s, Output) for s in outputs))
        self.assertSequenceEqual([s.name for s in outputs], list("123a"))

    def test_multi_input_summary(self):
        class Str(str):
            pass

        @summarize.register(Str)
        def _(s):
            return PartialSummary(str(s), None)

        class TestWidget(OWBaseWidget):
            class Inputs:
                input_a = MultiInput("A", Str)

            @Inputs.input_a
            def set_a(self, index, a):
                pass

            @Inputs.input_a.insert
            def insert_a(self, index, a):
                pass

            @Inputs.input_a.remove
            def remove_a(self, index):
                pass
        w = TestWidget()
        w.insert_a(0, Str("00"))
        w.insert_a(1, Str("11"))
        self.assertSequenceEqual(
            list(w.input_summaries["A"].values()),
            [PartialSummary("00", None), PartialSummary("11", None)])
        w.set_a(0, None)
        self.assertSequenceEqual(
            list(w.input_summaries["A"].values()),
            [PartialSummary("11", None)])
        w.set_a(0, Str("00"))
        self.assertSequenceEqual(
            list(w.input_summaries["A"].values()),
            [PartialSummary("00", None), PartialSummary("11", None)])
        w.insert_a(1, Str("05"))
        self.assertSequenceEqual(
            list(w.input_summaries["A"].values()),
            [PartialSummary("00", None), PartialSummary("05", None),
             PartialSummary("11", None)])
        w.set_a(1, None)
        w.remove_a(1)
        self.assertSequenceEqual(
            list(w.input_summaries["A"].values()),
            [PartialSummary("00", None), PartialSummary("11", None)])


if __name__ == "__main__":
    unittest.main()



# pylint: disable=all
from AnyQt.QtCore import Qt, QRect
from AnyQt.QtWidgets import QListView, QApplication, QProxyStyle, QStyle, QStyleFactory
from AnyQt.QtTest import QTest, QSignalSpy

from orangewidget.tests.base import GuiTest
from orangewidget.tests.utils import mouseMove, excepthook_catch

from orangewidget.utils import combobox


class StyleHintStyle(QProxyStyle):
    def styleHint(self, hint, option, widget=None, returnData=None) -> int:
        if hint == QStyle.SH_ComboBox_ListMouseTracking:
            return 1
        return super().styleHint(hint, option, widget, returnData)

    @staticmethod
    def create():
        return StyleHintStyle(QStyleFactory.create("fusion"))


class TestComboBoxSearch(GuiTest):
    def setUp(self):
        super().setUp()
        cb = combobox.ComboBoxSearch()
        cb.addItem("One")
        cb.addItem("Two")
        cb.addItem("Three")
        cb.insertSeparator(cb.count())
        cb.addItem("Four")
        self.cb = cb

    def tearDown(self):
        super().tearDown()
        self.cb.deleteLater()
        self.cb = None

    def test_combobox(self):
        cb = self.cb
        cb.grab()
        cb.showPopup()
        popup = cb.findChild(QListView)  # type: QListView
        # run through paint code for coverage
        popup.grab()
        cb.grab()

        model = popup.model()
        self.assertEqual(model.rowCount(), cb.count())
        QTest.keyClick(popup, Qt.Key_E)
        self.assertEqual(model.rowCount(), 2)
        QTest.keyClick(popup, Qt.Key_Backspace)
        self.assertEqual(model.rowCount(), cb.count())
        QTest.keyClick(popup, Qt.Key_F)
        self.assertEqual(model.rowCount(), 1)
        popup.setCurrentIndex(model.index(0, 0))
        spy = QSignalSpy(cb.activated[int])
        QTest.keyClick(popup, Qt.Key_Enter)

        self.assertEqual(spy[0], [4])
        self.assertEqual(cb.currentIndex(), 4)
        self.assertEqual(cb.currentText(), "Four")
        self.assertFalse(popup.isVisible())

    def test_combobox_navigation(self):
        cb = self.cb
        cb.setCurrentIndex(4)
        self.assertTrue(cb.currentText(), "Four")
        cb.showPopup()
        popup = cb.findChild(QListView)  # type: QListView
        self.assertEqual(popup.currentIndex().row(), 4)

        QTest.keyClick(popup, Qt.Key_Up)
        self.assertEqual(popup.currentIndex().row(), 2)
        QTest.keyClick(popup, Qt.Key_Escape)
        self.assertFalse(popup.isVisible())
        self.assertEqual(cb.currentIndex(), 4)
        cb.hidePopup()

    def test_click(self):
        interval = QApplication.doubleClickInterval()
        QApplication.setDoubleClickInterval(0)
        cb = self.cb
        spy = QSignalSpy(cb.activated[int])
        cb.showPopup()
        popup = cb.findChild(QListView)  # type: QListView
        model = popup.model()
        rect = popup.visualRect(model.index(2, 0))
        QTest.mouseRelease(
            popup.viewport(), Qt.LeftButton, Qt.NoModifier, rect.center()
        )
        QApplication.setDoubleClickInterval(interval)
        self.assertEqual(len(spy), 1)
        self.assertEqual(spy[0], [2])
        self.assertEqual(cb.currentIndex(), 2)

    def test_focus_out(self):
        cb = self.cb
        cb.showPopup()
        popup = cb.findChild(QListView)
        # Activate some other window to simulate focus out
        w = QListView()
        w.show()
        w.activateWindow()
        w.hide()
        self.assertFalse(popup.isVisible())

    def test_track(self):
        cb = self.cb
        cb.setStyle(StyleHintStyle.create())
        cb.showPopup()
        popup = cb.findChild(QListView)  # type: QListView
        model = popup.model()
        rect = popup.visualRect(model.index(2, 0))
        mouseMove(popup.viewport(), rect.center())
        self.assertEqual(popup.currentIndex().row(), 2)
        cb.hidePopup()

    def test_empty(self):
        cb = self.cb
        cb.clear()
        cb.showPopup()
        popup = cb.findChild(QListView)  # type: QListView
        self.assertIsNone(popup)

    def test_kwargs_enabled_focus_out(self):
        # PyQt5's property via kwargs can invoke virtual overrides while still
        # not fully constructed
        with excepthook_catch(raise_on_exit=True):
            combobox.ComboBoxSearch(enabled=False)

    def test_popup_util(self):
        geom = QRect(10, 10, 100, 400)
        screen = QRect(0, 0, 600, 600)
        g1 = combobox.dropdown_popup_geometry(
            geom, QRect(200, 100, 100, 20), screen
        )
        self.assertEqual(g1, QRect(200, 120, 100, 400))
        g2 = combobox.dropdown_popup_geometry(
            geom, QRect(-10, 0, 100, 20), screen
        )
        self.assertEqual(g2, QRect(0, 20, 100, 400))
        g3 = combobox.dropdown_popup_geometry(
            geom, QRect(590, 0, 100, 20), screen
        )
        self.assertEqual(g3, QRect(600 - 100, 20, 100, 400))
        g4 = combobox.dropdown_popup_geometry(
            geom, QRect(0, 500, 100, 20), screen
        )
        self.assertEqual(g4, QRect(0, 500 - 400, 100, 400))



import unittest
from orangewidget.settings import Setting, SettingProvider

SHOW_ZOOM_TOOLBAR = "show_zoom_toolbar"
SHOW_GRAPH = "show_graph"
GRAPH = "graph"
ZOOM_TOOLBAR = "zoom_toolbar"
SHOW_LABELS = "show_labels"
SHOW_X_AXIS = "show_x_axis"
SHOW_Y_AXIS = "show_y_axis"
ALLOW_ZOOMING = "allow_zooming"
A_LIST = "a_list"
A_SET = "a_set"
A_DICT = "a_dict"


class SettingProviderTestCase(unittest.TestCase):
    def setUp(self):
        global default_provider
        default_provider = SettingProvider(Widget)

    def tearDown(self):
        default_provider.settings[SHOW_GRAPH].default = True
        default_provider.settings[SHOW_ZOOM_TOOLBAR].default = True
        default_provider.providers[GRAPH].settings[SHOW_LABELS].default = True
        default_provider.providers[GRAPH].settings[SHOW_X_AXIS].default = True
        default_provider.providers[GRAPH].settings[SHOW_Y_AXIS].default = True
        default_provider.providers[ZOOM_TOOLBAR].settings[ALLOW_ZOOMING].default = True

    def test_registers_all_settings(self):
        self.assertDefaultSettingsEqual(default_provider, {
            GRAPH: {
                SHOW_LABELS: True,
                SHOW_X_AXIS: True,
                SHOW_Y_AXIS: True,
            },
            ZOOM_TOOLBAR: {
                ALLOW_ZOOMING: True,
            },
            SHOW_ZOOM_TOOLBAR: True,
            SHOW_GRAPH: True,
        })

    def test_initialize_sets_defaults(self):
        widget = Widget()

        self.assertEqual(widget.show_graph, True)
        self.assertEqual(widget.show_zoom_toolbar, True)
        self.assertEqual(widget.graph.a_list, [])
        self.assertEqual(widget.graph.a_set, {1, 2, 3})
        self.assertEqual(widget.graph.a_dict, {1: 2, 3: 4})
        self.assertEqual(widget.graph.show_labels, True)
        self.assertEqual(widget.graph.show_x_axis, True)
        self.assertEqual(widget.graph.show_y_axis, True)
        self.assertEqual(widget.zoom_toolbar.allow_zooming, True)

    def test_initialize_with_data_sets_values_from_data(self):
        widget = Widget()
        default_provider.initialize(widget, {
            SHOW_GRAPH: False,
            GRAPH: {
                SHOW_Y_AXIS: False
            }
        })

        self.assertEqual(widget.show_graph, False)
        self.assertEqual(widget.show_zoom_toolbar, True)
        self.assertEqual(widget.graph.show_labels, True)
        self.assertEqual(widget.graph.show_x_axis, True)
        self.assertEqual(widget.graph.show_y_axis, False)
        self.assertEqual(widget.zoom_toolbar.allow_zooming, True)

    def test_initialize_with_data_stores_initial_values_until_instance_is_connected(self):
        widget = Widget.__new__(Widget)

        default_provider.initialize(widget, {
            SHOW_GRAPH: False,
            GRAPH: {
                SHOW_Y_AXIS: False
            }
        })

        self.assertFalse(hasattr(widget.graph, SHOW_Y_AXIS))
        widget.graph = Graph()
        self.assertEqual(widget.graph.show_y_axis, False)

    def test_get_provider(self):
        self.assertEqual(default_provider.get_provider(BaseWidget), None)
        self.assertEqual(default_provider.get_provider(Widget),
                         default_provider)
        self.assertEqual(default_provider.get_provider(BaseGraph), None)
        self.assertEqual(default_provider.get_provider(Graph),
                         default_provider.providers[GRAPH])
        self.assertEqual(default_provider.get_provider(ExtendedGraph),
                         default_provider.providers[GRAPH])
        self.assertEqual(default_provider.get_provider(ZoomToolbar),
                         default_provider.providers[ZOOM_TOOLBAR])

    def test_pack_settings(self):
        widget = Widget()

        widget.show_graph = False
        widget.graph.show_y_axis = False

        packed_settings = default_provider.pack(widget)

        self.assertEqual(packed_settings, {
            SHOW_GRAPH: False,
            SHOW_ZOOM_TOOLBAR: True,
            GRAPH: {
                SHOW_LABELS: True,
                SHOW_X_AXIS: True,
                SHOW_Y_AXIS: False,
                A_LIST: [],
                A_SET: {1, 2, 3},
                A_DICT: {1: 2, 3: 4}
            },
            ZOOM_TOOLBAR: {
                ALLOW_ZOOMING: True,
            },
        })

    def test_unpack_settings(self):
        widget = Widget()
        default_provider.unpack(widget, {
            SHOW_GRAPH: False,
            GRAPH: {
                SHOW_Y_AXIS: False,
            },

        })

        self.assertEqual(widget.show_graph, False)
        self.assertEqual(widget.show_zoom_toolbar, True)
        self.assertEqual(widget.graph.show_labels, True)
        self.assertEqual(widget.graph.show_x_axis, True)
        self.assertEqual(widget.graph.show_y_axis, False)
        self.assertEqual(widget.zoom_toolbar.allow_zooming, True)

    def test_mutables_are_unpacked_in_place(self):
        widget = Widget()
        a_list = widget.graph.a_list
        a_set = widget.graph.a_set
        a_dict = widget.graph.a_dict
        default_provider.unpack(widget, {
            GRAPH: {
                A_LIST: [1, 2, 3],
                A_SET: {4, 5},
                A_DICT: {6: 7}
            },
        })
        self.assertIs(a_list, widget.graph.a_list)
        self.assertEqual(a_list, [1, 2, 3])
        self.assertIs(a_set, widget.graph.a_set)
        self.assertEqual(a_set, {4, 5})
        self.assertIs(a_dict, widget.graph.a_dict)
        self.assertEqual(a_dict, {6: 7})

    def test_traverse_settings_works_without_instance_or_data(self):
        settings = set()

        for setting, data, _ in default_provider.traverse_settings():
            settings.add(setting.name)

        self.assertEqual(settings, {
            SHOW_ZOOM_TOOLBAR, SHOW_GRAPH,
            SHOW_LABELS, SHOW_X_AXIS, SHOW_Y_AXIS, A_LIST, A_SET, A_DICT,
            ALLOW_ZOOMING})

    def test_traverse_settings_selects_correct_data(self):
        settings = {}
        graph_data = {SHOW_LABELS: 3, SHOW_X_AXIS: 4, SHOW_Y_AXIS: 5,
                      A_LIST: [], A_SET: {1, 2, 3}, A_DICT: {1: 2, 3: 4}}
        zoom_data = {ALLOW_ZOOMING: 6}
        all_data = {SHOW_GRAPH: 1,
                    SHOW_ZOOM_TOOLBAR: 2,
                    GRAPH: graph_data,
                    ZOOM_TOOLBAR: zoom_data}

        for setting, data, _ in default_provider.traverse_settings(all_data):
            settings[setting.name] = data

        self.assertEqual(
            settings,
            {
                SHOW_GRAPH: all_data,
                SHOW_ZOOM_TOOLBAR: all_data,
                SHOW_LABELS: graph_data,
                SHOW_X_AXIS: graph_data,
                SHOW_Y_AXIS: graph_data,
                A_LIST: graph_data,
                A_SET: graph_data,
                A_DICT: graph_data,
                ALLOW_ZOOMING: zoom_data,
            }
        )

    def test_traverse_settings_with_partial_data(self):
        settings = {}
        graph_data = {SHOW_LABELS: 3, SHOW_X_AXIS: 4}
        all_data = {SHOW_GRAPH: 1, SHOW_ZOOM_TOOLBAR: 2, GRAPH: graph_data}

        for setting, data, _ in default_provider.traverse_settings(all_data):
            settings[setting.name] = data

        self.assertEqual(
            settings,
            {
                SHOW_GRAPH: all_data,
                SHOW_ZOOM_TOOLBAR: all_data,
                SHOW_LABELS: graph_data,
                SHOW_X_AXIS: graph_data,
                SHOW_Y_AXIS: graph_data,
                A_LIST: graph_data,
                A_SET: graph_data,
                A_DICT: graph_data,
                ALLOW_ZOOMING: {},
            }
        )

    def test_traverse_settings_selects_correct_instance(self):
        settings = {}
        widget = Widget()

        for setting, _, instance in \
                default_provider.traverse_settings(instance=widget):
            settings[setting.name] = instance

        self.assertEqual(
            {
                SHOW_GRAPH: widget,
                SHOW_ZOOM_TOOLBAR: widget,
                SHOW_LABELS: widget.graph,
                SHOW_X_AXIS: widget.graph,
                SHOW_Y_AXIS: widget.graph,
                A_LIST: widget.graph,
                A_SET: widget.graph,
                A_DICT: widget.graph,
                ALLOW_ZOOMING: widget.zoom_toolbar,
            },
            settings
        )

    def test_traverse_settings_with_partial_instance(self):
        settings = {}
        widget = Widget()
        widget.graph = None

        for setting, _, instance in \
                default_provider.traverse_settings(instance=widget):
            settings[setting.name] = instance

        self.assertEqual(
            settings,
            {
                SHOW_GRAPH: widget,
                SHOW_ZOOM_TOOLBAR: widget,
                SHOW_LABELS: None,
                SHOW_X_AXIS: None,
                SHOW_Y_AXIS: None,
                A_LIST: None,
                A_SET: None,
                A_DICT: None,
                ALLOW_ZOOMING: widget.zoom_toolbar,
            }
        )

    def assertDefaultSettingsEqual(self, provider, defaults):
        for name, value in defaults.items():
            if isinstance(value, dict):
                self.assertIn(name, provider.providers)
                self.assertDefaultSettingsEqual(provider.providers[name], value)
            else:
                self.assertEqual(provider.settings[name].default, value)


def initialize_settings(instance):
    """This is usually done in Widget's new,
    but we avoid all that complications for tests."""
    provider = default_provider.get_provider(instance.__class__)
    if provider:
        provider.initialize(instance)
default_provider = None
""":type: SettingProvider"""


class BaseGraph:
    show_labels = Setting(True)

    def __init__(self):
        initialize_settings(self)


class Graph(BaseGraph):
    show_x_axis = Setting(True)
    show_y_axis = Setting(True)
    a_list = Setting([])
    a_set = Setting({1, 2, 3})
    a_dict = Setting({1: 2, 3: 4})

    def __init__(self):
        super().__init__()
        initialize_settings(self)


class ExtendedGraph(Graph):
    pass


class ZoomToolbar:
    allow_zooming = Setting(True)

    def __init__(self):
        initialize_settings(self)


class BaseWidget:
    settingsHandler = None

    show_graph = Setting(True)

    graph = SettingProvider(Graph)

    def __init__(self):
        initialize_settings(self)
        self.graph = Graph()


class Widget(BaseWidget):
    show_zoom_toolbar = Setting(True)

    zoom_toolbar = SettingProvider(ZoomToolbar)

    def __init__(self):
        super().__init__()
        initialize_settings(self)

        self.zoom_toolbar = ZoomToolbar()


if __name__ == '__main__':
    unittest.main(verbosity=2)
